/**
* This file is part of Simox.
*
* Simox is free software; you can redistribute it and/or modify
* it under the terms of the GNU Lesser General Public License as
* published by the Free Software Foundation; either version 2 of
* the License, or (at your option) any later version.
*
* Simox is distributed in the hope that it will be useful, but
* WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public License
* along with this program. If not, see <http://www.gnu.org/licenses/>.
*
* @package    VirtualRobot
* @author     Peter Kaiser
* @copyright  2015 Peter Kaiser
*             GNU Lesser General Public License
*
*/
#pragma once

#include "VirtualRobot/VirtualRobot.h"
#include "VirtualRobot/IK/JacobiProvider.h"

class SoSeparator;

namespace VirtualRobot
{
    class Constraint;
    typedef double (*OptimizationFunction)(std::vector<double> &gradient);

    class OptimizationFunctionSetup
    {
        public:
            unsigned int id;
            Constraint *constraint;
            bool soft;
    };

    class VIRTUAL_ROBOT_IMPORT_EXPORT Constraint : public JacobiProvider, public std::enable_shared_from_this<Constraint>
    {
    public:
        Constraint(const RobotNodeSetPtr& nodeSet);

        void initialize();

        virtual bool getRobotPoseForConstraint(Eigen::Matrix4f& pose);

        float getErrorDifference();

        const std::vector<OptimizationFunctionSetup> &getEqualityConstraints();
        const std::vector<OptimizationFunctionSetup> &getInequalityConstraints();
        const std::vector<OptimizationFunctionSetup> &getOptimizationFunctions();

        /*!
         * Each constraint implements its own optimization function that contributes to a combined
         * optimization function of a constrained IK problem.
         *
         * The individual optimization function value of a constraint is multiplied by the given factor
         * in order to weigh the contribution of the individual constraint.
         *
         * The default value of this factor is 1.
         */
        void setOptimizationFunctionFactor(float factor);
        float getOptimizationFunctionFactor();

        // Interface for NLopt-based solvers (default implementations)
        virtual double optimizationFunction(unsigned int id);
        virtual Eigen::VectorXf optimizationGradient(unsigned int id);

        // Interface for Jacobian-based solvers (default implementations)
        Eigen::MatrixXf getJacobianMatrix() override;
        Eigen::MatrixXf getJacobianMatrix(SceneObjectPtr tcp) override;
        Eigen::VectorXf getError(float stepSize = 1.0f) override;
        bool checkTolerances() override;

        virtual bool usingCollisionModel();

    protected:
        void addEqualityConstraint(unsigned int id, bool soft=false);
        void addInequalityConstraint(unsigned int id, bool soft=false);
        void addOptimizationFunction(unsigned int id, bool soft=false);

    protected:
        std::vector<OptimizationFunctionSetup> equalityConstraints;
        std::vector<OptimizationFunctionSetup> inequalityConstraints;
        std::vector<OptimizationFunctionSetup> optimizationFunctions;

        float lastError;
        float lastLastError;

        float optimizationFunctionFactor;
    };

    typedef std::shared_ptr<Constraint> ConstraintPtr;
}

